#!/usr/bin/env python

import numpy
import subprocess
import vtk
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt

class fluid:

    def __init__(self, name = 'unnamed'):
        '''
        A Navier-Stokes two-dimensional fluid flow simulation object. Most
        simulation values are assigned default values upon initialization.
        :param name: Simulation identifier
        :type name: str
        '''
        self.sim_id = name

        self.init_grid()
        self.current_time()
        self.end_time()
        self.file_time()
        self.safety_factor()
        self.max_iterations()
        self.tolerance_criteria()
        self.relaxation_parameter()
        self.upwind_differencing_factor()
        self.boundary_conditions()
        self.reynolds_number()
        self.gravity()

    def init_grid(self, nx = 10, ny = 10, dx = 0.1, dy = 0.1):
        '''
        Initializes the numerical grid.
        :param nx: Fluid grid width in number of cells
        :type nx: int
        :param ny: Fluid grid height in number of cells
        :type ny: int
        :param dx: Grid cell width (meters)
        :type dx: float
        :param dy: Grid cell height (meters)
        :type dy: float
        '''
        self.nx = numpy.asarray(nx)
        self.ny = numpy.asarray(ny)
        self.dx = numpy.asarray(dx)
        self.dy = numpy.asarray(dy)
        self.P = numpy.zeros((nx+2, ny+2))
        self.U = numpy.zeros((nx+2, ny+2))
        self.V = numpy.zeros((nx+2, ny+2))

    def current_time(self, t = 0.0):
        '''
        Set the current simulation time. Default value = 0.0.
        :param t: The current time value.
        :type t: float
        '''
        self.t = numpy.asarray(t)

    def end_time(self, t_end = 1.0):
        '''
        Set the simulation end time.
        :param t_end: The time when to stop the simulation.
        :type t_end: float
        '''
        self.t_end = numpy.asarray(t_end)

    def file_time(self, t_file = 0.1):
        '''
        Set the simulation output file interval.
        :param t_file: The time when to stop the simulation.
        :type t_file: float
        '''
        self.t_file = numpy.asarray(t_file)

    def safety_factor(self, tau = 0.5):
        '''
        Define the safety factor for the time step size control. Default value =
        0.5.
        :param tau: Safety factor in ]0;1]
        :type tau: float
        '''
        self.tau = numpy.asarray(tau)
        
    def max_iterations(self, itermax = 5000):
        '''
        Set the maximal allowed iterations per time step. Default value = 5000.
        :param itermax: Max. solution iterations in [1;inf[
        :type itermax: int
        '''
        self.itermax = numpy.asarray(itermax)

    def tolerance_criteria(self, epsilon = 1.0e-4):
        '''
        Set the tolerance criteria for the fluid solver. Default value = 1.0e-4.
        :param epsilon: Criteria value
        :type epsilon: float
        '''
        self.epsilon = numpy.asarray(epsilon)

    def relaxation_parameter(self, omega = 1.7):
        '''
        Set the relaxation parameter for the successive overrelaxation (SOR)
        solver. The solver is identical to the Gauss-Seidel method when omega =
        1. Default value = 1.7.
        :param omega: Relaxation parameter value, in ]0;2[
        :type omega: float
        '''
        self.omega = numpy.asarray(omega)

    def upwind_differencing_factor(self, gamma = 0.9):
        '''
        Set the upwind diffencing factor used in the finite difference
        approximations. Default value = 0.9.
        :param gamma: Upward differencing factor value, in ]0;1[
        :type gamma: float
        '''
        self.gamma = numpy.asarray(gamma)

    def boundary_conditions(self, left = 1, right = 1, top = 1, bottom = 1):
        '''
        Set the wall boundary conditions. The values correspond to the following
        conditions: 1) free-slip, 2) no-slip, 3) outflow, 4) periodic
        :param left, right, top, bottom: The wall to specify the BC for
        :type left, right, top, bottom: int
        '''
        self.w_left = numpy.asarray(left)
        self.w_right = numpy.asarray(right)
        self.w_top = numpy.asarray(top)
        self.w_bottom = numpy.asarray(bottom)

    def reynolds_number(self, re = 100):
        '''
        Define the simulation Reynolds number.
        :param re: Reynolds number in ]0;infty[
        :type re: float
        '''
        self.re = numpy.asarray(re)

    def gravity(self, gx = 0.0, gy = 0.0):
        '''
        Set the gravitational acceleration on the fluid.
        :param gx: Horizontal gravitational acceleration.
        :type gx: float
        :param gy: Vertical gravitational acceleration. Negative values are
            downward.
        :type gy: float
        '''
        self.gx = numpy.asarray(gx)
        self.gy = numpy.asarray(gy)

    def read(self, path, verbose = True):
        '''
        Read data file from disk.
        :param path: Path to data file
        :type path: str
        '''
        fh = None
        try:
            targetfile = path
            if verbose == True:
                print('Input file: ' + targetfile)
            fh = open(targetfile, 'rb')

            self.t      = numpy.fromfile(fh, dtype=numpy.float64, count=1)
            self.t_end  = numpy.fromfile(fh, dtype=numpy.float64, count=1)
            self.t_file = numpy.fromfile(fh, dtype=numpy.float64, count=1)
            self.tau    = numpy.fromfile(fh, dtype=numpy.float64, count=1)
            
            self.itermax = numpy.fromfile(fh, dtype=numpy.int32, count=1)
            self.epsilon = numpy.fromfile(fh, dtype=numpy.float64, count=1)
            self.omega   = numpy.fromfile(fh, dtype=numpy.float64, count=1)
            self.gamma   = numpy.fromfile(fh, dtype=numpy.float64, count=1)

            self.gx = numpy.fromfile(fh, dtype=numpy.float64, count=1)
            self.gy = numpy.fromfile(fh, dtype=numpy.float64, count=1)
            self.re = numpy.fromfile(fh, dtype=numpy.float64, count=1)

            self.w_left   = numpy.fromfile(fh, dtype=numpy.int32, count=1)
            self.w_right  = numpy.fromfile(fh, dtype=numpy.int32, count=1)
            self.w_top    = numpy.fromfile(fh, dtype=numpy.int32, count=1)
            self.w_bottom = numpy.fromfile(fh, dtype=numpy.int32, count=1)

            self.dx = numpy.fromfile(fh, dtype=numpy.float64, count=1)
            self.dy = numpy.fromfile(fh, dtype=numpy.float64, count=1)
            self.nx = numpy.fromfile(fh, dtype=numpy.int32, count=1)
            self.ny = numpy.fromfile(fh, dtype=numpy.int32, count=1)

            self.init_grid(dx = self.dx, dy = self.dy,\
                    nx = self.nx, ny = self.ny)

            for i in range(self.nx+2):
                for j in range(self.ny+2):
                    self.P[i,j] = \
                            numpy.fromfile(fh, dtype=numpy.float64, count=1)

            for i in range(self.nx+2):
                for j in range(self.ny+2):
                    self.U[i,j] = \
                            numpy.fromfile(fh, dtype=numpy.float64, count=1)

            for i in range(self.nx+2):
                for j in range(self.ny+2):
                    self.V[i,j] = \
                            numpy.fromfile(fh, dtype=numpy.float64, count=1)

        finally:
            if fh is not None:
                fh.close()

    def write(self, verbose = True, folder = './'):
        '''
        Write the simulation parameters to disk so that the fluid flow solver
        can read it.
        '''
        fh = None
        try:
            targetfile = folder + '/' + self.sim_id + '.dat'
            if verbose == True:
                print('Output file: ' + targetfile)
            fh = open(targetfile, 'wb')

            fh.write(self.t.astype(numpy.float64))
            fh.write(self.t_end.astype(numpy.float64))
            fh.write(self.t_file.astype(numpy.float64))
            fh.write(self.tau.astype(numpy.float64))
            
            fh.write(self.itermax.astype(numpy.int32))
            fh.write(self.epsilon.astype(numpy.float64))
            fh.write(self.omega.astype(numpy.float64))
            fh.write(self.gamma.astype(numpy.float64))

            fh.write(self.gx.astype(numpy.float64))
            fh.write(self.gy.astype(numpy.float64))
            fh.write(self.re.astype(numpy.float64))

            fh.write(self.w_left.astype(numpy.int32))
            fh.write(self.w_right.astype(numpy.int32))
            fh.write(self.w_top.astype(numpy.int32))
            fh.write(self.w_bottom.astype(numpy.int32))

            fh.write(self.dx.astype(numpy.float64))
            fh.write(self.dy.astype(numpy.float64))
            fh.write(self.nx.astype(numpy.int32))
            fh.write(self.ny.astype(numpy.int32))

            for i in range(self.nx+2):
                for j in range(self.ny+2):
                    fh.write(self.P[i,j].astype(numpy.float64))

            for i in range(self.nx+2):
                for j in range(self.ny+2):
                    fh.write(self.U[i,j].astype(numpy.float64))

            for i in range(self.nx+2):
                for j in range(self.ny+2):
                    fh.write(self.V[i,j].astype(numpy.float64))

        finally:
            if fh is not None:
                fh.close()

    def run(self):
        '''
        Run the simulation using the C program.
        '''
        self.write()
        subprocess.call('./ns2dfd ' + self.sim_id + '.dat', shell=True)

    def writeVTK(self, folder = './', verbose = True):
        '''
        Writes a VTK file for the fluid grid to the current folder by default.
        The file name will be in the format ``<self.sid>.vti``. The vti files
        can be used for visualizing the fluid in ParaView.

        The fluid grid is visualized by opening the vti files, and pressing
        "Apply" to import all fluid field properties. To visualize the scalar
        fields, such as the pressure, the porosity, the porosity change or the
        velocity magnitude, choose "Surface" or "Surface With Edges" as the
        "Representation". Choose the desired property as the "Coloring" field.
        It may be desirable to show the color bar by pressing the "Show" button,
        and "Rescale" to fit the color range limits to the current file. The
        coordinate system can be displayed by checking the "Show Axis" field.
        All adjustments by default require the "Apply" button to be pressed
        before regenerating the view.

        The fluid vector fields (e.g. the fluid velocity) can be visualizing by
        e.g. arrows. To do this, select the fluid data in the "Pipeline
        Browser". Press "Glyph" from the "Common" toolbar, or go to the
        "Filters" mennu, and press "Glyph" from the "Common" list. Make sure
        that "Arrow" is selected as the "Glyph type", and "Velocity" as the
        "Vectors" value. Adjust the "Maximum Number of Points" to be at least as
        big as the number of fluid cells in the grid. Press "Apply" to visualize
        the arrows.

        If several data files are generated for the same simulation (e.g. using
        the :func:`writeVTKall()` function), it is able to step the
        visualization through time by using the ParaView controls.

        :param folder: The folder where to place the output binary file (default
            (default = './')
        :type folder: str
        :param verbose: Show diagnostic information (default = True)
        :type verbose: bool
        '''
        filename = folder + '/' + self.sim_id + '.vti' # image grid

        # initalize VTK data structure
        grid = vtk.vtkImageData()
        grid.SetOrigin([0.0, 0.0, 0.0])
        grid.SetSpacing([self.dx, self.dy, 1])
        grid.SetDimensions([self.nx+2, self.ny+2, 1])

        # array of scalars: hydraulic pressures
        pres = vtk.vtkDoubleArray()
        pres.SetName("Pressure")
        pres.SetNumberOfComponents(1)
        pres.SetNumberOfTuples(grid.GetNumberOfPoints())

        # array of vectors: hydraulic velocities
        vel = vtk.vtkDoubleArray()
        vel.SetName("Velocity")
        vel.SetNumberOfComponents(2)
        vel.SetNumberOfTuples(grid.GetNumberOfPoints())

        # insert values
        for y in range(self.ny+2):
            for x in range(self.nx+2):
                    idx = x + (self.nx+2)*y
                    pres.SetValue(idx, self.P[x,y])
                    vel.SetTuple(idx, [self.U[x,y], self.V[x,y]])

        # add pres array to grid
        grid.GetPointData().AddArray(pres)
        grid.GetPointData().AddArray(vel)

        # write VTK XML image data file
        writer = vtk.vtkXMLImageDataWriter()
        writer.SetFileName(filename)
        writer.SetInput(grid)
        writer.Update()
        if (verbose == True):
            print('Output file: {0}'.format(filename))

    def plot_PUV(self, format = 'png'):
        plt.figure(figsize=[8,8])

        #ax = plt.subplot(1, 3, 1)
        plt.title("Pressure")
        imgplt = plt.imshow(self.P.T, origin='lower')
        imgplt.set_interpolation('nearest')
        #imgplt.set_interpolation('bicubic')
        #imgplt.set_cmap('hot')
        plt.xlabel('$x$')
        plt.ylabel('$y$')
        plt.colorbar()

        # show velocities as arrows
        Q = plt.quiver(self.U, self.V)

        # show velocities as stream lines
        #plt.streamplot(numpy.arange(self.nx+2),numpy.arange(self.ny+2),\
                #self.U, self.V)

        '''
        # show velocities as heat maps
        ax = plt.subplot(1, 3, 2)
        plt.title("U")
        imgplt = plt.imshow(self.U.T, origin='lower')
        imgplt.set_interpolation('nearest')
        #imgplt.set_interpolation('bicubic')
        #imgplt.set_cmap('hot')
        plt.xlabel('$x$')
        plt.ylabel('$y$')
        plt.colorbar()

        ax = plt.subplot(1, 3, 3)
        plt.title("V")
        imgplt = plt.imshow(self.V.T, origin='lower')
        imgplt.set_interpolation('nearest')
        #imgplt.set_interpolation('bicubic')
        #imgplt.set_cmap('hot')
        plt.xlabel('$x$')
        plt.ylabel('$y$')
        plt.colorbar()
        '''

        plt.savefig(self.sim_id + '-PUV.' + format, transparent=False)
