#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <string.h>
#include "file_io.h"
#include "utility.h"
#include "boundary.h"
#include "solution.h"

#define VERSION 0.1

/*#define DEBUG*/

int main(int argc, char** argv)
{
    double t, t_file, t_end, tau;
    int itermax;
    double epsilon, omega, gamma;
    double gx, gy, re;
    int w_left, w_right, w_top, w_bottom;
    double dx, dy;
    int nx, ny;
    double **P, **U, **V;
    double **F, **G, **RHS, **RES;

    double dt = 0.0;
    double res;
    long n = 0;
    int it = 0;
    int nfile = 0;
    double t_file_elapsed = 0.0;
    char filename[50];
    char *simulation_id;
    char *dot;

    int c;
    while ((c = getopt(argc, argv, "hv")) != -1)
        switch (c)
    {
        case 'h':
            printf("usage: %s [OPTIONS] FILENAME\n"
                    "where FILENAME is a valid input file.\n"
                    "-h\tDisplay help\n"
                    "-v\tDisplay version information\n", argv[0]);
            return 0;
            break;
        case 'v':
            printf("%s: A Navier Stokes solver in 2D using finite differences, "
                    "version %.1f\n"
                    "Written by Anders Damsgaard, "
                    "https://github.com/anders-dc/ns2dfd\n", argv[0], VERSION);
            return 0;
            break;
        case '?':
            if (isprint(optopt))
                fprintf(stderr, "Unknown option `-%c`.\n", optopt);
            else
                fprintf(stderr, "Unknown option character `\\x%x`.\n", optopt);
            return 1;

        default:
            abort();
    }

    if (optind == argc - 1) {
        read_file(argv[optind], &t, &t_end, &t_file, &tau, &itermax,
                &epsilon, &omega, &gamma, 
                &gx, &gy, &re, &w_left, &w_right, &w_top, &w_bottom,
                &dx, &dy, &nx, &ny, &P, &U, &V);
    } else {
        fprintf(stderr, "%s: no input file specified.\n", argv[0]);
        return 1;
    }

    simulation_id = argv[optind];
    dot = strchr(simulation_id, '.');
    dot[0] = '\0';
    printf("%s: simulation id: `%s`\n", argv[0], simulation_id);

    allocate_matrix(&F, nx+2, ny+2);
    allocate_matrix(&G, nx+2, ny+2);
    allocate_matrix(&RHS, nx+2, ny+2);
    allocate_matrix(&RES, nx+2, ny+2);

#ifdef DEBUG
    print_matrix("V", P, nx+2, ny+2);
    print_matrix("U", U, nx+2, ny+2);
    print_matrix("V", V, nx+2, ny+2);
#endif

    while (t <= t_end+dt) {

        dt = select_time_step(tau, re, dx, dy, nx, ny, U, V);

        if (t_file_elapsed >= t_file || n == 0) {
            sprintf(filename, "%s%05d.dat", simulation_id, nfile);
            write_file(filename, &t, &t_end, &t_file, &tau, &itermax,
                    &epsilon, &omega, &gamma, 
                    &gx, &gy, &re, &w_left, &w_right, &w_top, &w_bottom,
                    &dx, &dy, &nx, &ny, &P, &U, &V);
            t_file_elapsed = 0.0;
            nfile++;
        }

        /*printf("\rt = %f/%.2f s, dt = %f s, it = %d, last output: %d   ",*/
        printf("t = %f/%.2f s, dt = %f s, it = %d, last output: %d\n",
                t, t_end, dt, it, nfile-1);

        set_boundary_conditions(w_left, w_right, w_top, w_bottom, P, U, V,
                nx, ny);

        compute_f_g(F, G, U, V, P, re, nx, ny, dx, dy, gx, gy, dt, gamma);

#ifdef DEBUG
        print_matrix("F", F, nx+2, ny+2);
        print_matrix("G", G, nx+2, ny+2);
#endif

        compute_rhs(RHS, F, G, dt, dx, dy, nx, ny);

#ifdef DEBUG
        print_matrix("RHS", RHS, nx+2, ny+2);
#endif

        it = 0;
        res = 1.0e9;
        while ((it <= itermax) && (res >= epsilon)) {

            sor_cycle(P, RHS, omega, nx, ny, dx, dy);

            res = max_residual_norm(RES, P, RHS, nx, ny, dx, dy);

#ifdef DEBUG
            printf("\nit = %d\n", it);
            print_matrix("P", P, nx+2, ny+2);
            print_matrix("RES", RES, nx+2, ny+2);
#endif

            it++;
        }

        correct_velocities(U, V, F, G, P, nx, ny, dt, dx, dy);

#ifdef DEBUG
        print_matrix("U", U, nx+2, ny+2);
        print_matrix("V", V, nx+2, ny+2);
#endif

        t += dt;
        n++;
        t_file_elapsed += dt;
    }
    puts("\n");

    free_memory(P, U, V, nx);
    free_matrix(&F, nx+2);
    free_matrix(&G, nx+2);
    free_matrix(&RHS, nx+2);
    free_matrix(&RES, nx+2);
    return 0;
}
