#include <stdio.h>
#include <stdlib.h>
#include <math.h>

/* Second order central difference (d^2 u)/(dx^2) */
inline double ddu_dxx(double **U, int i, int j, double dx)
{
    return (U[i+1][j] - 2.0*U[i][j] + U[i-1][j])/(dx*dx);
}

/* Second order central difference (d^2 v)/(dx^2) */
inline double ddv_dxx(double **V, int i, int j, double dx)
{
    return (V[i+1][j] - 2.0*V[i][j] + V[i-1][j])/(dx*dx);
}

/* Second order central difference (d^2 u)/(dy^2) */
inline double ddu_dyy(double **U, int i, int j, double dy)
{
    return (U[i][j+1] - 2.0*U[i][j] + U[i][j-1])/(dy*dy);
}

/* Second order central difference (d^2 v)/(dy^2) */
inline double ddv_dyy(double **V, int i, int j, double dy)
{
    return (V[i][j+1] - 2.0*V[i][j] + V[i][j-1])/(dy*dy);
}

/* First order upwind difference (d(uu))/(dx) */
double duu_dx(double **U, int i, int j, double dx, double gamma)
{
    return 1.0/dx*(pow((U[i][j] + U[i+1][j])/2.0, 2.0) -
             pow((U[i-1][j] + U[i][j])/2.0, 2.0))
        + gamma/dx*(
                fabs(U[i][j] + U[i+1][j])/2.0 * (U[i][j] - U[i+1][j])/2.0 -
                fabs(U[i-1][j] + U[i][j])/2.0 * (U[i-1][j] - U[i][j])/2.0);
}

double dvv_dy(double **V, int i, int j, double dy, double gamma)
{
    return 1.0/dy*(pow((V[i][j] + V[i][j+1])/2.0, 2.0) -
            pow((V[i][j-1] + V[i][j])/2.0, 2.0))
        + gamma/dy*(
                fabs(V[i][j] + V[i][j+1])/2.0 * (V[i][j] - V[i][j+1])/2.0 -
                fabs(V[i][j-1] + V[i][j])/2.0 * (V[i][j-1] - V[i][j])/2.0);
}

/* First order upwind difference (d(uv))/(dy) */
double duv_dy(double **U, double **V, int i, int j, double dy, double gamma)
{
    return 1.0/dy*(
            (V[i][j] + V[i+1][j])/2.0 * (U[i][j] + U[i][j+1])/2.0 -
            (V[i][j-1] + V[i+1][j-1])/2.0 * (U[i][j-1] + U[i][j])/2.0)
        + gamma/dy*(
                fabs(V[i][j] + V[i+1][j])/2.0 * (U[i][j] - U[i][j+1])/2.0 -
                fabs(V[i][j-1] + V[i+1][j-1])/2.0 * (U[i][j-1] - U[i][j])/2.0);
}

double duv_dx(double **U, double **V, int i, int j, double dx, double gamma)
{
    return 1.0/dx*(
            (U[i][j] + U[i][j+1])/2.0 * (V[i][j] + V[i+1][j])/2.0 -
            (U[i-1][j] + U[i-1][j+1])/2.0 * (V[i-1][j] + V[i][j])/2.0)
        + gamma/dx*(
                fabs(U[i][j] + U[i][j+1])/2.0 * (V[i][j] - V[i+1][j])/2.0 -
                fabs(U[i-1][j] + U[i-1][j+1])/2.0 * (V[i-1][j] - V[i][j])/2.0);
}

/* Project new velocities based on Chorin's method, saved in F and G */
void compute_f_g(double **F, double **G,
        double **U, double **V, double **P, double re, 
        int nx, int ny, double dx, double dy,
        double gx, double gy, double dt, double gamma)
{
    int i, j;
    for (i=1; i<nx; i++)
        for (j=1; j<=ny; j++)
            F[i][j] = U[i][j]
                + dt*(1.0/re*(ddu_dxx(U, i, j, dx) + ddu_dyy(U, i, j, dy))
                        - duu_dx(U, i, j, dx, gamma)
                        - duv_dy(U, V, i, j, dy, gamma)
                        + gx);

    for (i=1; i<=nx; i++)
        for (j=1; j<ny; j++)
            G[i][j] = V[i][j]
                + dt*(1.0/re*(ddv_dxx(V, i, j, dx) + ddv_dyy(V, i, j, dy))
                        - duv_dx(U, V, i, j, dx, gamma)
                        - dvv_dy(V, i, j, dy, gamma)
                        + gy);

    for (j=1; j<=ny; j++) {
        F[0][j]  = U[0][j];
        F[nx][j] = U[nx][j];
    }

    for (i=1; i<=nx; i++) {
        G[i][0]  = V[i][0];
        G[i][ny] = V[i][ny];
    }
}

/* Find the right hand side value of the Poisson equation */
void compute_rhs(double **RHS, double **F, double **G,
        double dt, double dx, double dy, int nx, int ny)
{
    int i, j;
    for (i=1; i<=nx; i++) {
        for (j=1; j<=ny; j++) {
            RHS[i][j] = 1.0/dt*(
                    (F[i][j] - F[i-1][j])/dx + (G[i][j] - G[i][j-1])/dy);
        }
    }
}

void e_parameters(double *e_left, double *e_right,
        double *e_bottom, double *e_top,
        int i, int j, int nx, int ny)
{
    if (i == 1) {
        *e_left = 0.0;
    } else if (i > 1) {
        *e_left = 1.0;
    } else {
        fprintf(stderr, "e_parameters: i value %d not understood\n", i);
        exit(1);
    }

    if (i < nx) {
        *e_right = 1.0;
    } else if (i == nx) {
        *e_right = 0.0;
    } else {
        fprintf(stderr, "e_parameters: i value %d not understood\n", i);
        exit(1);
    }

    if (j == 1) {
        *e_bottom = 0.0;
    } else if (j > 1) {
        *e_bottom = 1.0;
    } else {
        fprintf(stderr, "e_parameters: j value %d not understood\n", j);
        exit(1);
    }

    if (j < ny) {
        *e_top = 1.0;
    } else if (j == ny) {
        *e_top = 0.0;
    } else {
        fprintf(stderr, "e_parameters: j value %d not understood\n", j);
        exit(1);
    }
}

/* Compute a successive overrelaxation (SOR) cycle */
void sor_cycle(double **P, double **RHS, double omega,
        int nx, int ny, double dx, double dy)
{
    int i, j;
    double e_left, e_right, e_top, e_bottom;

    for (i=1; i<=nx; i++) {
        for (j=1; j<=ny; j++) {
            e_parameters(&e_left, &e_right, &e_bottom, &e_top, i, j, nx, ny);
            P[i][j] = (1.0 - omega)*P[i][j] +
                omega/((e_right + e_left)/(dx*dx) + (e_top + e_bottom)/(dy*dy))
                *( (e_right*P[i+1][j] + e_left*P[i-1][j])/(dx*dx) +
                        (e_top*P[i][j+1] + e_bottom*P[i][j-1])/(dy*dy) -
                        RHS[i][j]);
        }
    }
}

void calculate_residuals(double **RES, double **P, double **RHS,
        int nx, int ny, double dx, double dy)
{
    int i, j;
    double e_left, e_right, e_top, e_bottom;

    for (i=1; i<=nx; i++) {
        for (j=1; j<=ny; j++) {
            e_parameters(&e_left, &e_right, &e_bottom, &e_top, i, j, nx, ny);
            RES[i][j] =
                (e_right*(P[i+1][j] - P[i][j]) - e_left*(P[i][j] - P[i-1][j]))
                /(dx*dx) +
                (e_top*(P[i][j+1] - P[i][j]) - e_bottom*(P[i][j] - P[i][j-1]))
                /(dy*dy) - RHS[i][j];
            if (RES[i][j] != RES[i][j]) {
                fprintf(stderr, "calculate_residuals: "
                        "Error, residual is %f in cell i=%d, j=%d\n",
                        RES[i][j], i, j);
                exit(1);
            }
        }
    }
}

double max_residual_norm(double **RES, double **P, double **RHS,
        int nx, int ny, double dx, double dy)
{
    int i, j;
    double val;
    double max = -1.0e9;

    calculate_residuals(RES, P, RHS, nx, ny, dx, dy);

    for (i=1; i<=nx; i++) {
        for (j=1; j<=ny; j++) {
            val = fabs(RES[i][j]);
            if (val > max)
                max = val;
        }
    }
    return max;
}

void correct_velocities(double **U, double **V, double **F, double **G, 
        double **P, int nx, int ny, double dt, double dx, double dy)
{
    int i, j;
    for (i=1; i<nx; i++)
        for (j=1; j<=ny; j++)
            U[i][j] = F[i][j] - dt/dx*(P[i+1][j] - P[i][j]);

    for (i=1; i<=nx; i++)
        for (j=1; j<ny; j++)
            V[i][j] = G[i][j] - dt/dy*(P[i][j+1] - P[i][j]);
}
