#include <stdio.h>
#include <stdlib.h>
#include <math.h>

int allocate_matrix(double ***M, int nx, int ny)
{
    int i;
    (*M) = (double**) malloc(sizeof(double*)*nx);
    if (*M == NULL)
        return 1;

    for (i=0; i<nx; i++) {
        (*M)[i] = (double*) malloc(sizeof(double)*ny);
        if ((*M)[i] == NULL)
            return 1;
    }
    return 0;
}

void free_matrix(double ***M, int nx)
{
    int i;
    for (i=0; i<nx; i++)
        free((*M)[i]);
    free(*M);
}

int allocate_memory(double ***P, double ***U, double ***V, int nx, int ny)
{
    if (allocate_matrix(P, nx+2, ny+2) != 0) {
        fprintf(stderr, "allocate_memory: Could not allocate memory for P\n");
        return 1;
    }
    if (allocate_matrix(U, nx+2, ny+2) != 0) {
        fprintf(stderr, "allocate_memory: Could not allocate memory for U\n");
        return 1;
    }
    if (allocate_matrix(V, nx+2, ny+2) != 0) {
        fprintf(stderr, "allocate_memory: Could not allocate memory for V\n");
        return 1;
    }
    return 0;
}

void free_memory(double **P, double **U, double **V, int nx)
{
    free_matrix(&P, nx+2);
    free_matrix(&U, nx+2);
    free_matrix(&V, nx+2);
}

double max_abs_value(double **M, int nx, int ny)
{
    int i, j;
    double tmp;
    double max = -1.0e19;
    for (j=1; j<=ny; j++) {
        for (i=1; i<=nx; i++) {
            tmp = fabs(M[i][j]);
            if (tmp > max) max = tmp;
        }
    }
    return max;
}

double select_time_step(double tau, double re, double dx, double dy,
        int nx, int ny, double **U, double **V)
{
    double t_diff, t_adv_u, t_adv_v;
    double u_max, v_max;
    double dt;
    u_max = max_abs_value(U, nx, ny);
    v_max = max_abs_value(V, nx, ny);

    t_diff = re/2.0*(1.0/(1.0/(dx*dx) + (1.0/(dy*dy))));
    t_adv_u = dx/(u_max + 1.0e-12);
    t_adv_v = dx/(v_max + 1.0e-12);

    dt = fmin(t_diff, fmin(t_adv_u, t_adv_v));

    if (dt < 1.0e-12) {
        fprintf(stderr, "select_time_step: Error, the time step is too small: "
                "%f s.\n"
                "tau = %f, Re = %f, u_max = %f, v_max = %f\n"
                "t_diff = %f s, t_adv_u = %f s, t_adv_v = %f s\n",
                dt, tau, re, u_max, v_max, t_diff, t_adv_u, t_adv_v);
        exit(1);
    }
    return dt;
}

void print_matrix(char* description, double **M, int nx, int ny)
{
    int i, j;
    printf("%s:\n", description);
    for (j=0; j<ny; j++) {
        for (i=0; i<nx; i++) {
            printf("%.2f\t", M[i][j]);
        }
        puts("\n");
    }
}
